// Winamp test visualization library v1.3
// Copyright (C) 1997-1998, Justin Frankel/Nullsoft
// Feel free to base any plugins on this "framework"...
//
// Fixed by Sebastian Pipping <webmaster@hartwork.org>
//
//
// 1.3
//        No-black-border mode fixed
// 1.21
//        Solid oscilliscope module added
//        Thick spectrum module added
//        Alternative spectrum module added
//        Window position now saved on WM_DESTROY not WM_MOVE
// 1.1   
//        GetWindowRect bug fixed
//        Full array visualization
// 1.0



#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "vis.h"



#define NO_BLACK_BORDER



char szAppName[] = "SimpleVis"; // Window class, etc


// Configuration declarations
int config_x = 50, config_y = 50; // Screen X position and Y position, respectively
void config_read( struct winampVisModule * this_mod );   // Reads the configuration
void config_write( struct winampVisModule * this_mod );  // Writes the configuration
void config_getinifn( struct winampVisModule * this_mod, char * ini_file ); // Makes the INI file filename

// Returns a winampVisModule when requested. Used in hdr, below
winampVisModule * getModule( int which );

// "Member" functions
void config( struct winampVisModule * this_mod );      // Configuration dialog
int init( struct winampVisModule * this_mod );         // Initialization for module
int render_osc( struct winampVisModule * this_mod );   // Rendering for osc module
int render_spec( struct winampVisModule * this_mod );  // Rendering for spec module
int render_vu( struct winampVisModule * this_mod );    // Rendering for vu module
void quit( struct winampVisModule * this_mod );        // Deinitialization for module

// Window procedure
LRESULT CALLBACK WndProc( HWND hwnd, UINT message, WPARAM wp, LPARAM lp );
HWND hMainWnd; // main window handle

// Double buffering data
HDC memDC;      // Memory device context
HBITMAP	memBM;  // Memory bitmap (for memDC)
HBITMAP	oldBM;  // Old bitmap (from memDC)


// Module header, includes version, description, and address
// of the module retriever function
winampVisHeader hdr = {
	VIS_HDRVER,
	"Nullsoft Test Visualization Library v1.3",
	getModule
};


winampVisModule mod_osc_lines =
{
	"Oscilliscope (lines)",
	NULL,	// hwndParent
	NULL,	// hDllInstance
	0,		// sRate
	0,		// nCh
	25,		// latencyMS
	25,		// delayMS
	0,		// spectrumNch
	2,		// waveformNch
	{ 0, },	// spectrumData
	{ 0, },	// waveformData
	config,
	init,
	render_osc, 
	quit
};


winampVisModule mod_osc_solid =
{
	"Oscilliscope (solid)",
	NULL,	// hwndParent
	NULL,	// hDllInstance
	0,		// sRate
	0,		// nCh
	25,		// latencyMS
	25,		// delayMS
	0,		// spectrumNch
	2,		// waveformNch
	{ 0, },	// spectrumData
	{ 0, },	// waveformData
	config,
	init,
	render_osc, 
	quit
};


winampVisModule mod_spec_thin =
{
	"Spectrum Analyser (thin)",
	NULL,	// hwndParent
	NULL,	// hDllInstance
	0,		// sRate
	0,		// nCh
	25,		// latencyMS
	25,		// delayMS
	2,		// spectrumNch
	0,		// waveformNch
	{ 0, },	// spectrumData
	{ 0, },	// waveformData
	config,
	init,
	render_spec, 
	quit
};


winampVisModule mod_spec_thick =
{
	"Spectrum Analyser (thick)",
	NULL,	// hwndParent
	NULL,	// hDllInstance
	0,		// sRate
	0,		// nCh
	25,		// latencyMS
	25,		// delayMS
	2,		// spectrumNch
	0,		// waveformNch
	{ 0, },	// spectrumData
	{ 0, },	// waveformData
	config,
	init,
	render_spec, 
	quit
};


winampVisModule mod_spec_alt =
{
	"Spectrum Analyser (alternative)",
	NULL,	// hwndParent
	NULL,	// hDllInstance
	0,		// sRate
	0,		// nCh
	25,		// latencyMS
	25,		// delayMS
	2,		// spectrumNch
	0,		// waveformNch
	{ 0, },	// spectrumData
	{ 0, },	// waveformData
	config,
	init,
	render_spec, 
	quit
};



winampVisModule mod_vu =
{
	"VU Meter",
	NULL,	// hwndParent
	NULL,	// hDllInstance
	0,		// sRate
	0,		// nCh
	25,		// latencyMS
	25,		// delayMS
	0,		// spectrumNch
	2,		// waveformNch
	{ 0, },	// spectrumData
	{ 0, },	// waveformData
	config,
	init,
	render_vu, 
	quit
};


// This is the only exported symbol. returns our main header.
// If you are compiling C++, the extern "C" { is necessary, so we just #ifdef it
#ifdef __cplusplus
extern "C" {
#endif
__declspec( dllexport ) winampVisHeader * winampVisGetHeader()
{
	return &hdr;
}
#ifdef __cplusplus
}
#endif


// Getmodule routine from the main header. Returns NULL if an invalid module
// was requested, otherwise returns either mod_osc, mod_spec or mod_vu depending
// on <which>.
winampVisModule * getModule( int which )
{
	switch( which )
	{
	case 0:  return &mod_osc_lines;
	case 1:  return &mod_osc_solid;
	case 2:  return &mod_spec_thin;
	case 3:  return &mod_spec_thick;
	case 4:  return &mod_spec_alt;
	case 5:  return &mod_vu;
	default: return NULL;
	}
}


// Configuration. Passed this_mod, as a "this" parameter.
// Allows you to make one configuration function that shares code for all your
// modules (you don't HAVE to use it though, you can make config1(), config2(),
// etc...)
void config( struct winampVisModule * this_mod )
{
	MessageBox(
		this_mod->hwndParent,
		"This module is Copyright (c) 1997-1998, Justin Frankel/Nullsoft  \n"
			"-- This is just a demonstration module, it really isn't\n"
			"   supposed to be enjoyable --\n"
			"\n"
			"Fixed by Sebastian Pipping <webmaster@hartwork.org>",
		"About",
		MB_OK
	);
}


// Initialization. Registers the window class, creates the window, etc.
// Again, this one works for both modules, but you could make init1() and init2()...
// returns 0 on success, 1 on failure.
int init( struct winampVisModule * this_mod )
{
	// Width and height are the same for mod_osc and mod_spec,
	// but mod_vu is shaped differently
	int width   = ( this_mod == &mod_vu ) ? 256 : 576;  
	int height  = ( this_mod == &mod_vu ) ?  32 : 256;  
	config_read( this_mod );

	{
		// Register our window class
		WNDCLASS wc;
		memset( &wc, 0, sizeof( WNDCLASS ) );
		wc.lpfnWndProc    = WndProc;                        // Window procedure
		wc.hInstance      = this_mod->hDllInstance;         // Instance of DLL
		wc.hCursor        = LoadCursor( NULL, IDC_ARROW );  // Default cursor
		wc.lpszClassName  = szAppName;                      // Window class name

		if( !RegisterClass( &wc ) )
		{
			MessageBox(
				this_mod->hwndParent,
				"Error registering window class",
				"blah",
				MB_OK
			);
			return 1;
		}
	}


	hMainWnd = CreateWindowEx(
		WS_EX_TOOLWINDOW |        // These exstyles put a nice small frame
			WS_EX_APPWINDOW,      // but also a button in the taskbar
		szAppName,                // Window class name
		this_mod->description,    // Use description for a window title
		WS_VISIBLE | WS_SYSMENU,  // Make the window visible with a close button
		config_x,                 // Screen position (read from config)
		config_y,                 //
		width,                    // Initial width
		height,                   // Initial height
		this_mod->hwndParent,     // Parent window (winamp main window)
		NULL,                     // No menu
		this_mod->hDllInstance,   // Instance of DLL
		0                         // No window creation data
	);
	
	if( !hMainWnd )
	{
		MessageBox(
			this_mod->hwndParent,
			"Error creating window",
			"blah",
			MB_OK
		);
		return 1;
	}

	// Set user data to a "this" pointer
	SetWindowLong( hMainWnd, GWL_USERDATA, ( LONG )this_mod );

	{
		// Adjust size of window to make the client area exactly width*height
		RECT r;
		GetClientRect( hMainWnd, &r );
		SetWindowPos(
			hMainWnd,
			NULL,
			0,
			0,
			width  * 2 - r.right,
			height * 2 - r.bottom,
			SWP_NOMOVE | SWP_NOZORDER
		);
	}

	// Create doublebuffer
	memDC = CreateCompatibleDC( NULL );
	memBM = CreateCompatibleBitmap( memDC, width, height );
	oldBM = ( HBITMAP )SelectObject( memDC, memBM );

	// Show window
	ShowWindow( hMainWnd, SW_SHOWNORMAL );
	return 0;
}


// Render function for oscilliscope. Returns 0 if successful, 1 if visualization should end.
int render_osc( struct winampVisModule * this_mod )
{
	int x, y;
	const int shift_by = this_mod->nCh - 1;

	// Clear background
#ifdef NO_BLACK_BORDER
	const RECT r = { 0, 0, 576, 256 };
	FillRect( memDC, &r, GetStockObject( WHITE_BRUSH ) );
#else
	Rectangle( memDC, 0, 0, 576, 256 );
#endif

	// Draw oscilliscope
	for( y = 0; y < this_mod->nCh; y++ )
	{
		if( this_mod == &mod_osc_lines )
		{
			// Lines (original)
			MoveToEx( memDC, 0, ( y * 256 + this_mod->waveformData[ y ][ 0 ] ^ 128 ) >> shift_by, NULL );
			for( x = 1; x < 576; x++ )
			{
				LineTo( memDC, x, ( y * 256 + this_mod->waveformData[ y ][ x ] ^ 128 ) >> shift_by );
			}
		}
		else
		{
			// Solid
			for( x = 0; x < 576; x++ )
			{
				MoveToEx( memDC, x, ( y * 256 + 127 ) >> shift_by, NULL );
				LineTo( memDC, x, ( y * 256 + this_mod->waveformData[ y ][ x ] ^ 128 ) >> shift_by );
			}
		}
	}

	{
		// Copy doublebuffer to window
		HDC hdc = GetDC( hMainWnd );
		BitBlt( hdc, 0, 0, 576, 256, memDC, 0, 0, SRCCOPY );
		ReleaseDC( hMainWnd, hdc );
	}
	return 0;
}


// Render function for analyser. Returns 0 if successful, 1 if visualization should end.
int render_spec( struct winampVisModule * this_mod )
{
	int x, y;
	const int shift_by = this_mod->nCh - 1;

	// Clear background
#ifdef NO_BLACK_BORDER
	const RECT r = { 0, 0, 576, 256 };
	FillRect( memDC, &r, GetStockObject( WHITE_BRUSH ) );
#else
	Rectangle( memDC, 0, 0, 576, 256 );
#endif
	
	// Draw analyser
	for( y = 0; y < this_mod->nCh; y++ )
	{
		if( this_mod == &mod_spec_thin )
		{
			// Thin (original)
			for( x = 0; x < 576; x++ )
			{
				MoveToEx( memDC, x, ( y * 256 + 256 ) >> shift_by, NULL );
				LineTo( memDC, x, ( y * 256 + 256 - this_mod->spectrumData[ y ][ x ] ) >> shift_by );
			}
		}
		else if( this_mod == &mod_spec_thick )
		{
			// Thick
			for( x = 0; x < 576; x += 3 )
			{
				int max = 0;
				int i;
				for( i = 0; i < 3; i++ )
				{
					if( this_mod->spectrumData[ y ][ x + i ] > max )
					{
						max = this_mod->spectrumData[ y ][ x + i ];	
					}
				}
				if( max == 0 ) max = 1;
				
				MoveToEx( memDC, x    , ( y * 256 + 256 ) >> shift_by, NULL );
				LineTo(   memDC, x    , ( y * 256 + 256 - max ) >> shift_by );
				MoveToEx( memDC, x + 1, ( y * 256 + 256 ) >> shift_by, NULL );
				LineTo(   memDC, x + 1, ( y * 256 + 256 - max ) >> shift_by );
				MoveToEx( memDC, x + 2, ( y * 256 + 256 ) >> shift_by, NULL );
				LineTo(   memDC, x + 2, ( y * 256 + 256 - max ) >> shift_by );
			}
		}
		else
		{
			// Alternative
			if( ( y & 1 ) == 0 )
			{
				// Upper
				MoveToEx( memDC, 0, ( y * 256 + 256 - this_mod->spectrumData[ y ][ 0 ] ) >> shift_by, NULL );
				for( x = 2; x < 576; x += 2 )
				{
					LineTo( memDC, x, ( y * 256 + 256 - this_mod->spectrumData[ y ][ x ] ) >> shift_by );
				}
				LineTo( memDC, x, ( y * 256 + 256 - this_mod->spectrumData[ y ][ 575 ] ) >> shift_by );
			}
			else
			{
				// Lower
				MoveToEx( memDC, 0, ( y * 256 + this_mod->spectrumData[ y ][ 0 ] ) >> shift_by, NULL );
				for( x = 2; x < 576; x += 2 )
				{
					LineTo( memDC, x, ( y * 256 + this_mod->spectrumData[ y ][ x ] ) >> shift_by );
				}
				LineTo( memDC, x, ( y * 256 + this_mod->spectrumData[ y ][ 575 ] ) >> shift_by );
			}
		}
	}

	{
		// Copy doublebuffer to window
		HDC hdc = GetDC( hMainWnd );
		BitBlt( hdc, 0, 0, 576, 256, memDC, 0, 0, SRCCOPY );
		ReleaseDC( hMainWnd, hdc );
	}
	return 0;
}


// Render function for VU meter. Returns 0 if successful, 1 if visualization should end.
int render_vu( struct winampVisModule * this_mod )
{
	int x, y;
	
	// Clear background
#ifdef NO_BLACK_BORDER
	const RECT r = { 0, 0, 256, 32 };
	FillRect( memDC, &r, GetStockObject( WHITE_BRUSH ) );
#else
	Rectangle( memDC, 0, 0, 256, 32 );
#endif

	// Draw center
	MoveToEx( memDC, 128, 0, NULL );
	LineTo( memDC, 128, 32 );

	// Draw VU meter
	for( y = 0; y < 2; y++ )
	{
		int last = this_mod->waveformData[ y ][ 0 ];
		int total = 0;
		for( x = 1; x < 576; x++ )
		{
			total += abs( last - this_mod->waveformData[ y ][ x ] );
			last = this_mod->waveformData[ y ][ x ];
		}

		total /= 576;
		if( total > 127 ) total = 127;
		if( y )
		{
			MoveToEx( memDC, 128 + total, 0, NULL );
			LineTo( memDC, 128 + total, 32 );
		}
		else
		{
			MoveToEx( memDC, 128 - total, 0, NULL );
			LineTo( memDC, 128 - total, 32 );
		}
	}

	{
		// Copy doublebuffer to window
		HDC hdc = GetDC( hMainWnd );
		BitBlt( hdc, 0, 0, 256, 32, memDC, 0, 0, SRCCOPY );
		ReleaseDC( hMainWnd, hdc );
	}
	return 0;
}


// Cleanup (opposite of init()). Destroys the window, unregisters the window class
void quit( struct winampVisModule * this_mod )
{
	// Delete window
	DestroyWindow( hMainWnd );

	// Write configuration
	config_write( this_mod );
	
	// Delete doublebuffer
	SelectObject( memDC, oldBM );
	DeleteObject( memDC );
	DeleteObject( memBM );
	
	// Unregister window class
	UnregisterClass( szAppName, this_mod->hDllInstance );
}


// Window procedure
LRESULT CALLBACK WndProc( HWND hwnd, UINT message, WPARAM wp, LPARAM lp )
{
	switch (message)
	{
	case WM_CREATE:
	case WM_ERASEBKGND:
		return 0;

	case WM_PAINT:
		{
			// Update from doublebuffer
			PAINTSTRUCT ps;
			RECT r;
			HDC hdc = BeginPaint( hwnd, &ps );
			GetClientRect( hwnd, &r );
			BitBlt( hdc, 0, 0, r.right, r.bottom, memDC, 0, 0, SRCCOPY );
			EndPaint( hwnd, &ps );
			return 0;
		}

	case WM_KEYDOWN:
	case WM_KEYUP:
		{
			// Pass keyboard messages to main winamp window
			// Get <this_mod> from window's user data
			winampVisModule * this_mod = ( winampVisModule * )GetWindowLong( hwnd, GWL_USERDATA );
			PostMessage( this_mod->hwndParent, message, wp, lp );
			return 0;
		}

	case WM_DESTROY:
		{
			// Note:  Window still exists

			// Update <config_x> and <config_y>
			RECT r;

			// Note:  GetWindowRect can fail, that's where
			//        the negative values came from
			if( GetWindowRect( hMainWnd, &r ) )
			{
				config_x = r.left;
				config_y = r.top;
			}

			// Stop message loop
			PostQuitMessage( 0 );
			return 0;
		}
	}
	return DefWindowProc( hwnd, message, wp, lp );
}


void config_getinifn( struct winampVisModule * this_mod, char * ini_file )
{
	// Makes "plugin.ini" in the directory this DLL is stored in
	char * p;
	GetModuleFileName( this_mod->hDllInstance, ini_file, MAX_PATH );
	p = ini_file + strlen( ini_file );
	while( ( p >= ini_file ) && ( *p != '\\' ) ) p--;
	if( ++p >= ini_file ) *p = '\0';
	strcpy( p, "plugin.ini" );
}


void config_read( struct winampVisModule * this_mod )
{
	char ini_file[ MAX_PATH ];
	config_getinifn( this_mod, ini_file );
	config_x = GetPrivateProfileInt( this_mod->description, "Screen_x", config_x, ini_file );
	config_y = GetPrivateProfileInt( this_mod->description, "Screen_y", config_y, ini_file );

	// Ensure valid values
	int width   = ( this_mod == &mod_vu ) ? 256 : 576;
	int height  = ( this_mod == &mod_vu ) ?  32 : 256;
	if(	( config_x < 0 ) ||
		( config_x > ( GetSystemMetrics( SM_CXVIRTUALSCREEN ) - width ) ) ||
		( config_y < 0 ) ||
		( config_y > ( GetSystemMetrics( SM_CYVIRTUALSCREEN ) - height ) ) )
	{
		config_x = 50;
		config_y = 50;
	}
}


void config_write( struct winampVisModule * this_mod )
{
	char string[ 32 ];
	char ini_file[ MAX_PATH ];
	config_getinifn( this_mod, ini_file );

	wsprintf( string, "%u", config_x );
	WritePrivateProfileString( this_mod->description, "Screen_x", string, ini_file );
	wsprintf( string, "%u", config_y );
	WritePrivateProfileString( this_mod->description, "Screen_y", string, ini_file );
}
